#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
#include <stdio.h>	/* stderr declared */
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined, if not ANSI-C */
#include "medi.h"
/*----------------------------------------------------------------------*/
#define	IMAX	(101*101)	/* maximum number of data points */
#define	ABS(x)	(((x) < 0)? -(x): (x))
#define	MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define	MIN(a,b)	(((a) < (b)) ? (a) : (b))
/*----------------------------------------------------------------------
	Finds the Median, or any other Quantile.
	The median of n float items a[i] is defined as the item which is
	less than (or equal to) half of the n items and which is larger
	than (or equal to) the other half of the n items. The method to
	be displayed easily generalizes to the problem of finding the
	kth smallest of n items. Finding the median represents the
	special case k = n/2.
	The algorithm invented by C. A. R. Hoare functions as
	follows. First, the partitioning operation of Quicksort (p. 77)
	is applied with l = 0 and r = (n-1), and with a[k] selected as
	splitting value x. The resulting index values i and j are such that
		1. a[h] <= x for all h < i
		2. a[h] >= x for all h > j
		3. i > j
	There are three possible cases that may arise:
		1. The splitting value x was too small; as a result,
		   the limit between the two partitions is below the
		   desired value k. The partitioning process has to be
		   repeated upon the elements a[i]...a[r].
		2. The chosen bound x was too large. The splitting
		   operation has to be repeated on the partition
		   a[l]...a[j].
		3. j < k < i: the element a[k] splits the array into two
		   partitions in the specified proportions and
		   therefore is the desired quantile.
	The splitting process has to be repeated until case 3. arises.
------------------------------------------------------------------------
	Niklaus Wirth:	Algorithms + Data Structure = Programs, p. 82.
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static short Find_short(int n, short a[], int k)
#else	/* old style */
static short Find_short(n, a, k)
int n, k;
short a[];
#endif
{
	int l, r, i, j;
	short w, x;

	l = 0;
	r = n - 1;

	while (l < r) {
		x = a[k];
		i = l;
		j = r;
		/* begin split */
		while (i <= j) {
			while (a[i] < x)
				i++;
			while (x < a[j])
				j--;
			if (i <= j) {
				w = a[i];
				a[i] = a[j];
				a[j] = w;
				i++;
				j--;
			}
		}
		/* end split */
		if (j < k)
			l = i;
		if (k < i)
			r = j;
	}

	return a[k];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static LONG Find_long(int n, LONG a[], int k)
#else	/* old style */
static LONG Find_long(n, a, k)
int n, k;
LONG a[];
#endif
{
	int l, r, i, j;
	LONG w, x;

	l = 0;
	r = n - 1;

	while (l < r) {
		x = a[k];
		i = l;
		j = r;
		/* begin split */
		while (i <= j) {
			while (a[i] < x)
				i++;
			while (x < a[j])
				j--;
			if (i <= j) {
				w = a[i];
				a[i] = a[j];
				a[j] = w;
				i++;
				j--;
			}
		}
		/* end split */
		if (j < k)
			l = i;
		if (k < i)
			r = j;
	}

	return a[k];
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static float Find_float(int n, float a[], int k)
#else	/* old style */
static float Find_float(n, a, k)
int n, k;
float a[];
#endif
{
	int l, r, i, j;
	float w, x;

	l = 0;
	r = n - 1;

	while (l < r) {
		x = a[k];
		i = l;
		j = r;
		/* begin split */
		while (i <= j) {
			while (a[i] < x)
				i++;
			while (x < a[j])
				j--;
			if (i <= j) {
				w = a[i];
				a[i] = a[j];
				a[j] = w;
				i++;
				j--;
			}
		}
		/* end split */
		if (j < k)
			l = i;
		if (k < i)
			r = j;
	}

	return a[k];
}
/*------------------------------------------------------------------------
   medi_uchar() calculates a trimmed and weighted Mean of a set of images.
   It uses the Median Absolute Deviation (MAD) from the Median to select
   pixels to be omitted from the Mean. The MAD of a Normal Distribution
   is 0.67 times the usual Standard Deviation. Pixels are omitted if they
   deviate more than limit*MAD. The plain median is used if limit == 0.0.
   The weights used in calculating the final trimmed mean are given in the
   float array weight[].
--------------------------------------------------------------------------
   Bjarne Thomsen, Aarhus, 27-Jul-92, 29-Dec-92, 28-Jun-93, 15-jul-93,
                           22-Aug-98
------------------------------------------------------------------------*/
#if	defined(ANSI)
void medi_uchar(UCHAR buf[],
const UCHAR *arr[],
const float weight[],
const int nimg,
const int npix,
const double limit)
#else	/* old style */
void medi_uchar(buf, arr, weight, nimg, npix, limit)
UCHAR buf[];
const UCHAR *arr[];
const float weight[];
const int nimg;
const int npix;
const double limit;
#endif
{
	int i, j, k1, k2, n;
	short a[IMAX], w[IMAX], med, mad, min, max, hw;
	float mean, wsum;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;

	for (j = 0; j < npix; ++j) {	/* take the j'th pixel */

		/* store pixel values from images in a[] and w[] */
		n = 0;
		for (i = 0; i < nimg; ++i) {
			w[i] = (a[i] = arr[i][j]);
			++n;
		}

		med = 0;	/* in case n == 0 */
		if (n > 2) {	/* 3 or more images */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_short(n, w, k1);
			if (k2 != k1) {
				med += Find_short(n, w, k2);
				med /= 2;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_short(n, w, k1);
				if(k2 != k1) {
					mad += Find_short(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (short) (limit*mad + 0.5);
				min = med - hw;
				max = med + hw;

				/* find the trimmed and weighted mean value */
				wsum = Zero;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += weight[i] * (float)a[i];
						wsum += weight[i];
					}
				mean /= wsum;
				med = (short) (mean + Half);
			}
		} else if (n > 0) {	/* 1 or 2 images */
			wsum = Zero;
			mean = Zero;
			for (i = 0; i < n; ++i) {
				mean += weight[i] * (float)a[i];
				wsum += weight[i];
			}
			mean /= wsum;
			med = (short) (mean + Half);
		}
		buf[j] = (UCHAR) med;
	}
}
/*------------------------------------------------------------------------
	rmed_uchar() calculates a running trimmed Mean for a given row.
	It uses the Median Absolute Deviation (MAD) from the Median to select
	pixels to be omitted from the Mean. The MAD of a Normal Distribution
	is 0.67 times the usual Standard Deviation. Pixels are omitted if they
	deviate more than limit*MAD. The plain median is used if limit == 0.0
	It returns (-1) if the running box is too large, otherwise (0).
--------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 12-July-93, 15-jul-93, 27-Nov-93
------------------------------------------------------------------------*/
#if	defined(ANSI)
int rmed_uchar(UCHAR buf[],
const int y,
const UCHAR *arr[],
const int nrows,
const int ncols,
const int my,
const int mx,
const double limit)
#else	/* old style */
int rmed_uchar(buf, y, arr, nrows, ncols, my, mx, limit)
UCHAR buf[];
const int y;
const UCHAR *arr[];
const int nrows;
const int ncols;
const int my;
const int mx;
const double limit;
#endif
{
	int x, i, j, k1, k2, npts, n;
	int nx, ny, xmin, xmax, ymin, ymax;
	short a[IMAX], w[IMAX], med, mad, min, max, hw;
	float mean;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;

	if ((2*my+1)*(2*mx+1) > IMAX) {
		fprintf(stderr, "rmed_uchar: Area of running box > %d\n", IMAX);
		return -1;
	}

	ymin = MAX(-1,y-my);
	ymax = MIN(nrows,y+my);
	ny = MIN(y-ymin,ymax-y);
	ymin = MAX(0,y-ny);
	ymax = MIN(nrows-1,y+ny);
	for (x = 0; x < ncols; ++x) {	/* select box centred on pixel (x,y) */
		xmin = MAX(-1,x-mx);
		xmax = MIN(ncols,x+mx);
		nx = MIN(x-xmin,xmax-x);
		xmin = MAX(0,x-nx);
		xmax = MIN(ncols-1,x+nx);

		/* store pixel values from running box in a[] and w[] */
		n = 0;
		for (j = ymin; j <= ymax; ++j)
			for (i = xmin; i <= xmax; ++i) {
				w[n] = (a[n] = arr[j][i]);
				++n;
			}

		med = 0;	/* in case n == 0 */
		if (n > 2) {	/* 3 or more data points */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_short(n, w, k1);
			if (k2 != k1) {
				med += Find_short(n, w, k2);
				med /= 2;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_short(n, w, k1);
				if(k2 != k1) {
					mad += Find_short(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (short) (limit*mad + 0.5);
				min = med - hw;
				max = med + hw;

				/* find the trimmed mean value */
				npts = 0;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += (float) a[i];
						++npts;
					}
				mean /= (float) npts;
				med = (short) (mean + Half);
			}
		} else if (n > 0) {	/* 1 or 2 data points */
			mean = Zero;
			for (i = 0; i < n; ++i)
				mean += (float) a[i];
			mean /= (float) n;
			med = (short) (mean + Half);
		}
		buf[x] = (UCHAR) med;
	}
	return 0;
}
/*------------------------------------------------------------------------
   medi_short() calculates a trimmed and weighted Mean of a set of images.
   It uses the Median Absolute Deviation (MAD) from the Median to select
   pixels to be omitted from the Mean. The MAD of a Normal Distribution
   is 0.67 times the usual Standard Deviation. Pixels are omitted if they
   deviate more than limit*MAD. The plain median is used if limit == 0.0.
   The weights used in calculating the final trimmed mean are given in the
   float array weight[].
--------------------------------------------------------------------------
   Bjarne Thomsen, Aarhus, 27-Jul-92, 29-Dec-92, 28-Jun-93, 15-Jul-93,
                           22-Aug-98
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void medi_short(short buf[],
const short *arr[],
const float weight[],
const int nimg,
const int npix,
const double limit,
const double blank)
#else	/* old style */
void medi_short(buf, arr, weight, nimg, npix, limit, blank)
short buf[];
const short *arr[];
const float weight[];
const int nimg;
const int npix;
const double limit;
const double blank;
#endif
{
	int i, j, k1, k2, n;
	short a[IMAX], w[IMAX], med, mad, min, max;
	LONG lmin, lmax, hw;
	float mean, wsum;
	const short b = (short) blank;
	const LONG ShortMax = SHORT_Max;
	const LONG ShortMin = SHORT_Min;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;

	for (j = 0; j < npix; ++j) {	/* take the j'th pixel */

		/* store valid pixel values from images in a[] and w[] */
		n = 0;
		for (i = 0; i < nimg; ++i)
			if (arr[i][j] != b) {
				w[n] = (a[n] = arr[i][j]);
				++n;
			}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid images */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_short(n, w, k1);
			if (k2 != k1) {
				LONG lmed = (LONG) med;	/* prevent overflow */
				lmed += (LONG) Find_short(n, w, k2);
				lmed /= 2L;
				med = (short) lmed;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_short(n, w, k1);
				if(k2 != k1) {
					mad += Find_short(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (LONG) (limit*mad + 0.5);
				lmin = (LONG) med - hw;
				lmax = (LONG) med + hw;

				min = (short) MAX(lmin,ShortMin);
				max = (short) MIN(lmax,ShortMax);

				/* find the trimmed and weighted mean value */
				wsum = Zero;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += weight[i] * (float)a[i];
						wsum += weight[i];
					}
				mean /= wsum;
				med = (short) ((mean < Zero)? mean-Half: mean+Half);
			}
		} else if (n > 0) {	/* 1 or 2 images */
			wsum = Zero;
			mean = Zero;
			for (i = 0; i < n; ++i) {
				mean += weight[i] * (float)a[i];
				wsum += weight[i];
			}
			mean /= wsum;
			med = (short) ((mean < Zero)? mean-Half: mean+Half);
		}
		buf[j] = med;
	}
}
/*------------------------------------------------------------------------
	rmed_short() calculates a running trimmed Mean for a given row.
	It uses the Median Absolute Deviation (MAD) from the Median to select
	pixels to be omitted from the Mean. The MAD of a Normal Distribution
	is 0.67 times the usual Standard Deviation. Pixels are omitted if they
	deviate more than limit*MAD. The plain median is used if limit == 0.0
	It returns (-1) if the running box is too large, otherwise (0).
--------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 12-Jul-93, 15-Jul-93, 27-Nov-93
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int rmed_short(short buf[],
const int y,
const short *arr[],
const int nrows,
const int ncols,
const int my,
const int mx,
const double limit,
const double blank)
#else	/* old style */
int rmed_short(buf, y, arr, nrows, ncols, my, mx, limit, blank)
short buf[];
const int y;
const short *arr[];
const int nrows;
const int ncols;
const int my;
const int mx;
const double limit;
const double blank;
#endif
{
	int x, i, j, k1, k2, n, npts;
	int nx, ny, xmin, xmax, ymin, ymax;
	short a[IMAX], w[IMAX], med, mad, min, max, hw;
	float mean;
	const short b = (short) blank;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;

	if ((2*my+1)*(2*mx+1) > IMAX) {
		fprintf(stderr, "rmed_short: Area of running box > %d\n", IMAX);
		return -1;
	}

	ymin = MAX(-1,y-my);
	ymax = MIN(nrows,y+my);
	ny = MIN(y-ymin,ymax-y);
	ymin = MAX(0,y-ny);
	ymax = MIN(nrows-1,y+ny);
	for (x = 0; x < ncols; ++x) {	/* select box centred on pixel (x,y) */
		xmin = MAX(-1,x-mx);
		xmax = MIN(ncols,x+mx);
		nx = MIN(x-xmin,xmax-x);
		xmin = MAX(0,x-nx);
		xmax = MIN(ncols-1,x+nx);

		/* store valid pixels from running box in a[] and w[] */
		n = 0;
		for (j = ymin; j <= ymax; ++j)
			for (i = xmin; i <= xmax; ++i)
				if (arr[j][i] != b) {
					w[n] = (a[n] = arr[j][i]);
					++n;
				}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid data points */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_short(n, w, k1);
			if (k2 != k1) {
				LONG lmed = (LONG) med;	/* prevent overflow */
				lmed += (LONG) Find_short(n, w, k2);
				lmed /= 2L;
				med = (short) lmed;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_short(n, w, k1);
				if(k2 != k1) {
					mad += Find_short(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (short) (limit*mad + 0.5);
				min = med - hw;
				max = med + hw;

				/* find the trimmed mean value */
				npts = 0;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += (float) a[i];
						++npts;
					}
				mean /= (float) npts;
				med = (short) ((mean < Zero)? mean-Half: mean+Half);
			}
		} else if (n > 0) {	/* 1 or 2 data points */
			mean = Zero;
			for (i = 0; i < n; ++i)
				mean += (float) a[i];
			mean /= (float) n;
			med = (short) ((mean < Zero)? mean-Half: mean+Half);
		}
		buf[x] = med;
	}
	return 0;
}
/*------------------------------------------------------------------------
   medi_long() calculates a trimmed and weighted Mean of a set of images.
   It uses the Median Absolute Deviation (MAD) from the Median to select
   pixels to be omitted from the Mean. The MAD of a Normal Distribution
   is 0.67 times the usual Standard Deviation. Pixels are omitted if they
   deviate more than limit*MAD. The plain median is used if limit == 0.0.
   The weights used in calculating the final trimmed mean are given in
   the float array weight[].
--------------------------------------------------------------------------
   Bjarne Thomsen, Aarhus, 27-Jul-92, 29-Dec-92, 28-Jun-93, 15-Jul-93,
                           22-Aug-98
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void medi_long(LONG buf[],
const LONG *arr[],
const float weight[],
const int nimg,
const int npix,
const double limit,
const double blank)
#else	/* old style */
void medi_long(buf, arr, weight, nimg, npix, limit, blank)
LONG buf[];
const LONG *arr[];
const float weight[];
const int nimg;
const int npix;
const double limit;
const double blank;
#endif
{
	int i, j, k1, k2, n;
	LONG a[IMAX], w[IMAX], med, mad, min, max, hw;
	float mean, wsum;
	const LONG b = (LONG) blank;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;
	const float Two = (float) 2;

	for (j = 0; j < npix; ++j) {	/* take the j'th pixel */

		/* store valid pixel values from images in a[] and w[] */
		n = 0;
		for (i = 0; i < nimg; ++i)
			if (arr[i][j] != b) {
				w[n] = (a[n] = arr[i][j]);
				++n;
			}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid images */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_long(n, w, k1);
			if (k2 != k1) {
				float fmed = (float) med; /* prevent overflow */
				fmed += (float) Find_long(n, w, k2);
				fmed /= Two;
				med = (LONG) fmed;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_long(n, w, k1);
				if(k2 != k1) {
					mad += Find_long(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (LONG) (limit*mad + 0.5);
				min = med - hw;
				max = med + hw;

				/* find the trimmed and weighted mean value */
				wsum = Zero;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += weight[i] * (float)a[i];
						wsum += weight[i];
					}
				mean /= wsum;
				med = (LONG) ((mean < Zero)? mean-Half: mean+Half);
			}
		} else if (n > 0) {	/* 1 or 2 images */
			wsum = Zero;
			mean = Zero;
			for (i = 0; i < n; ++i) {
				mean += weight[i] * (float)a[i];
				wsum += weight[i];
			}
			mean /= wsum;
			med = (LONG) ((mean < Zero)? mean-Half: mean+Half);
		}
		buf[j] = med;
	}
}
/*------------------------------------------------------------------------
	rmed_long() calculates a running trimmed Mean for a given row.
	It uses the Median Absolute Deviation (MAD) from the Median to select
	pixels to be omitted from the Mean. The MAD of a Normal Distribution
	is 0.67 times the usual Standard Deviation. Pixels are omitted if they
	deviate more than limit*MAD. The plain median is used if limit == 0.0
	It returns (-1) if the running box is too large, otherwise (0).
--------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 12-Jul-93, 15-Jul-93, 27-Nov-93
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int rmed_long(LONG buf[],
const int y,
const LONG *arr[],
const int nrows,
const int ncols,
const int my,
const int mx,
const double limit,
const double blank)
#else	/* old style */
int rmed_long(buf, y, arr, nrows, ncols, my, mx, limit, blank)
LONG buf[];
const int y;
const LONG *arr[];
const int nrows;
const int ncols;
const int my;
const int mx;
const double limit;
const double blank;
#endif
{
	int x, i, j, k1, k2, n, npts;
	int nx, ny, xmin, xmax, ymin, ymax;
	LONG a[IMAX], w[IMAX], med, mad, min, max, hw;
	float mean;
	const LONG b = (LONG) blank;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;
	const float Two = (float) 2;

	if ((2*my+1)*(2*mx+1) > IMAX) {
		fprintf(stderr, "rmed_long: Area of running box > %d\n", IMAX);
		return -1;
	}

	ymin = MAX(-1,y-my);
	ymax = MIN(nrows,y+my);
	ny = MIN(y-ymin,ymax-y);
	ymin = MAX(0,y-ny);
	ymax = MIN(nrows-1,y+ny);
	for (x = 0; x < ncols; ++x) {	/* select box centred on pixel (x,y) */
		xmin = MAX(-1,x-mx);
		xmax = MIN(ncols,x+mx);
		nx = MIN(x-xmin,xmax-x);
		xmin = MAX(0,x-nx);
		xmax = MIN(ncols-1,x+nx);

		/* store valid pixels from running box in a[] and w[] */
		n = 0;
		for (j = ymin; j <= ymax; ++j)
			for (i = xmin; i <= xmax; ++i)
				if (arr[j][i] != b) {
					w[n] = (a[n] = arr[j][i]);
					++n;
				}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid data points */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_long(n, w, k1);
			if (k2 != k1) {
				float fmed = (float) med;	/* prevent overflow */
				fmed += (float) Find_long(n, w, k2);
				fmed /= Two;
				med = (LONG) fmed;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_long(n, w, k1);
				if(k2 != k1) {
					mad += Find_long(n, w, k2);
					++mad;
					mad >>= 1;
				}
				hw = (LONG) (limit*mad + 0.5);
				min = med - hw;
				max = med + hw;

				/* find the trimmed mean value */
				npts = 0;
				mean = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						mean += (float) a[i];
						++npts;
					}
				mean /= (float) npts;
				med = (LONG) ((mean < Zero)? mean-Half: mean+Half);
			}
		} else if (n > 0) {	/* 1 or 2 data points */
			mean = Zero;
			for (i = 0; i < n; ++i)
				mean += (float) a[i];
			mean /= (float) n;
			med = (LONG) ((mean < Zero)? mean-Half: mean+Half);
		}
		buf[x] = med;
	}
	return 0;
}
/*------------------------------------------------------------------------
   fmedian() calculates the plain median of a float array a[].
   It is found by using Find_float() on a working copy w[] of a[].
--------------------------------------------------------------------------
   Bjarne Thomsen, Aarhus, 26-Oct-98
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
float fmedian(const int n, const float a[], float w[])
#else	/* old style */
float fmedian(n, a, w)
const int n;
const float a[];
float w[];
#endif
{
	int i, k1, k2;
	float med;
	const float Zero = (float) 0;
	const float Two = (float) 2;

	for (i = 0; i < n; ++i)
		w[i] = a[i];

	if (n > 2) {         /* 3 or more array elements */

		/* find the Median of w[] */
		/* note that w[] is reordered */
		k1 = (n - 1) >> 1;
		k2 = n >> 1;
		med = Find_float(n, w, k1);
		if (k2 != k1) {
			med += Find_float(n, w, k2);
			med /= Two;
		}

	} else if (n > 0) {  /* 1 or 2 array elements */

		med  = Zero;
		for (i = 0; i < n; ++i)
			med  += w[i];
		med /= (float) n;

	} else               /* no array elements */

		med = Zero;

	return med;
}
/*------------------------------------------------------------------------
   medi_float() calculates a trimmed and weighted Mean of a set of images.
   It uses the Median Absolute Deviation (MAD) from the Median to select
   pixels to be omitted from the Mean. The MAD of a Normal Distribution
   is 0.67 times the usual Standard Deviation. Pixels are omitted if they
   deviate more than limit*MAD. The plain median is used if limit == 0.0.
   The weights used in calculating the final trimmed mean are given in
   the float array weight[].
--------------------------------------------------------------------------
   Bjarne Thomsen, Aarhus, 27-Jul-92, 29-Dec-92, 28-Jun-93, 15-Jul-93,
                           08-Jun-98, 22-Aug-98
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void medi_float(float buf[],
const float *arr[],
const float weight[],
const int nimg,
const int npix,
const double limit,
const double blank)
#else	/* old style */
void medi_float(buf, arr, weight, nimg, npix, limit, blank)
float buf[];
const float *arr[];
const float weight[];
const int nimg;
const int npix;
const double limit;
const double blank;
#endif
{
	int i, j, k1, k2, n;
	float a[IMAX], w[IMAX], wsum, med, mad, min, max, hw;
	const float b = (float) blank;
	const float h = (float) (0.9*blank);
	const float Zero = (float) 0;
	const float Two = (float) 2;

	for (j = 0; j < npix; ++j) {	/* take the j'th pixel */

		/* store valid pixel values from images in a[] and w[] */
		n = 0;
		for (i = 0; i < nimg; ++i)
			if (arr[i][j] <= h) {
				w[n] = (a[n] = arr[i][j]);
				++n;
			}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid images */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_float(n, w, k1);
			if (k2 != k1) {
				med += Find_float(n, w, k2);
				med /= Two;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_float(n, w, k1);
				if(k2 != k1) {
					mad += Find_float(n, w, k2);
					mad /= Two;
				}
				hw = (float) (limit*mad);
				min = med - hw;
				max = med + hw;

				/* find the trimmed and weighted mean value */
				wsum = Zero;
				med = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						med  += weight[i] * a[i];
						wsum += weight[i];
					}
				med /= wsum;
			}
		} else if (n > 0) {	/* 1 or 2 images */
			wsum = Zero;
			med  = Zero;
			for (i = 0; i < n; ++i) {
				med  += weight[i] * a[i];
				wsum += weight[i];
			}
			med /= wsum;
		}
		buf[j] = med;
	}
}
/*------------------------------------------------------------------------
	rmed_float() calculates a running trimmed Mean for a given row.
	It uses the Median Absolute Deviation (MAD) from the Median to select
	pixels to be omitted from the Mean. The MAD of a Normal Distribution
	is 0.67 times the usual Standard Deviation. Pixels are omitted if they
	deviate more than limit*MAD. The plain median is used if limit == 0.0
	It returns (-1) if the running box is too large, otherwise (0).
--------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 12-Jul-93, 15-Jul-93, 27-Nov-93, 08-Jun-98
------------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int rmed_float(float buf[],
const int y,
const float *arr[],
const int nrows,
const int ncols,
const int my,
const int mx,
const double limit,
const double blank)
#else	/* old style */
int rmed_float(buf, y, arr, nrows, ncols, my, mx, limit, blank)
float buf[];
const int y;
const float *arr[];
const int nrows;
const int ncols;
const int my;
const int mx;
const double limit;
const double blank;
#endif
{
	int x, i, j, k1, k2, n, npts;
	int nx, ny, xmin, xmax, ymin, ymax;
	float a[IMAX], w[IMAX], med, mad, min, max, hw;
	const float b = (float) blank;
	const float h = (float) (0.9*blank);
	const float Zero = (float) 0;
	const float Two = (float) 2;

	if ((2*my+1)*(2*mx+1) > IMAX) {
		fprintf(stderr, "rmed_float: Area of running box > %d\n", IMAX);
		return -1;
	}

	ymin = MAX(-1,y-my);
	ymax = MIN(nrows,y+my);
	ny = MIN(y-ymin,ymax-y);
	ymin = MAX(0,y-ny);
	ymax = MIN(nrows-1,y+ny);
	for (x = 0; x < ncols; ++x) {	/* select box centred on pixel (x,y) */
		xmin = MAX(-1,x-mx);
		xmax = MIN(ncols,x+mx);
		nx = MIN(x-xmin,xmax-x);
		xmin = MAX(0,x-nx);
		xmax = MIN(ncols-1,x+nx);

		/* store valid pixels from running box in a[] and w[] */
		n = 0;
		for (j = ymin; j <= ymax; ++j)
			for (i = xmin; i <= xmax; ++i)
				if (arr[j][i] <= h) {
					w[n] = (a[n] = arr[j][i]);
					++n;
				}

		med = b;	/* blank in case n == 0 */
		if (n > 2) {	/* 3 or more valid data points */

			/* find the Median of w[] */
			/* note that w[] is reordered */
			k1 = (n - 1) >> 1;
			k2 = n >> 1;
			med = Find_float(n, w, k1);
			if (k2 != k1) {
				med += Find_float(n, w, k2);
				med /= Two;
			}

			if (limit > 0.0) {	/* calculate the trimmed mean */

				/* find the Absolute Deviations of w[] */
				for (i = 0; i < n; ++i)
					w[i] -= med;
				for (i = 0; i < n; ++i)
					w[i] = ABS(w[i]);

				/* find the Median of the Absolute Deviations */
				/* note that w[] is reordered */
				mad = Find_float(n, w, k1);
				if(k2 != k1) {
					mad += Find_float(n, w, k2);
					mad /= Two;
				}
				hw = (float) (limit*mad);
				min = med - hw;
				max = med + hw;

				/* find the trimmed mean value */
				npts = 0;
				med = Zero;
				for (i = 0; i < n; ++i)
					if (min <= a[i] && a[i] <= max) {
						med += a[i];
						++npts;
					}
				med /= (float) npts;
			}
		} else if (n > 0) {	/* 1 or 2 data points */
			med = Zero;
			for (i = 0; i < n; ++i)
				med += a[i];
			med /= (float) n;
		}
		buf[x] = med;
	}
	return 0;
}
